@@ -1,13 +1,17 @@
 from __future__ import unicode_literals
 
 import re
-from collections import OrderedDict
+import typing as T
 
 from conllu.exceptions import ParseException
-from conllu.models import Metadata, Token
+from conllu.models import Metadata, Token, TokenList
+
+_IdType = T.Union[int, T.Tuple[int, str, int]]
+_FieldParserType = T.Callable[[T.List[str], int], T.Any]
+_MetadataParserType = T.Callable[[str, T.Optional[str]], T.Any]
 
 DEFAULT_FIELDS = ('id', 'form', 'lemma', 'upos', 'xpos', 'feats', 'head', 'deprel', 'deps', 'misc')
-DEFAULT_FIELD_PARSERS = {
+DEFAULT_FIELD_PARSERS: T.Dict[str, _FieldParserType] = {
     "id": lambda line, i: parse_id_value(line[i]),
     "xpos": lambda line, i: parse_nullable_value(line[i]),
     "feats": lambda line, i: parse_dict_value(line[i]),
@@ -15,12 +19,14 @@ DEFAULT_FIELD_PARSERS = {
     "deps": lambda line, i: parse_paired_list_value(line[i]),
     "misc": lambda line, i: parse_dict_value(line[i]),
 }
-DEFAULT_METADATA_PARSERS = {
+DEFAULT_METADATA_PARSERS: T.Dict[str, _MetadataParserType] = {
     "newpar": lambda key, value: (key, value),
     "newdoc": lambda key, value: (key, value),
 }
 
-def parse_conllu_plus_fields(in_file, metadata_parsers=None):
+def parse_conllu_plus_fields(in_file: T.TextIO,
+                             metadata_parsers: T.Optional[T.Dict[str, _MetadataParserType]] = None
+                             ) -> T.Optional[T.Sequence[str]]:
     pos = in_file.tell()
 
     # Get first line
@@ -34,7 +40,7 @@ def parse_conllu_plus_fields(in_file, metadata_parsers=None):
     in_file.seek(pos)
 
     if not first_line.startswith("#"):
-        return
+        return None
 
     _, metadata = parse_token_and_metadata(first_line, metadata_parsers=metadata_parsers)
 
@@ -44,8 +50,8 @@ def parse_conllu_plus_fields(in_file, metadata_parsers=None):
 
     return fields
 
-def parse_sentences(in_file):
-    buf = []
+def parse_sentences(in_file: T.TextIO) -> T.Iterator[str]:
+    buf: T.List[str] = []
     for line in in_file:
         if line == "\n":
             if not buf:
@@ -57,7 +63,10 @@ def parse_sentences(in_file):
     if buf:
         yield "".join(buf).rstrip()
 
-def parse_token_and_metadata(data, fields=None, field_parsers=None, metadata_parsers=None):
+def parse_token_and_metadata(data: str, fields: T.Optional[T.Sequence[str]] = None,
+                             field_parsers: T.Optional[T.Dict[str, _FieldParserType]] = None,
+                             metadata_parsers: T.Optional[T.Dict[str, _MetadataParserType]] = None
+                             ) -> T.Tuple[T.List[Token], Metadata]:
     if not data:
         raise ParseException("Can't create TokenList, no data sent to constructor.")
 
@@ -88,7 +97,9 @@ def parse_token_and_metadata(data, fields=None, field_parsers=None, metadata_par
 
     return tokens, metadata
 
-def parse_line(line, fields, field_parsers=None):
+def parse_line(line: str,
+               fields: T.Sequence[str], field_parsers: T.Optional[T.Dict[str, _FieldParserType]] = None
+               ) -> Token:
     # Be backwards compatible if people called parse_line without field_parsers before
     field_parsers = field_parsers or DEFAULT_FIELD_PARSERS
 
@@ -103,32 +114,34 @@ def parse_line(line, fields, field_parsers=None):
     if "upos" not in field_parsers and "upostag" in field_parsers:
         field_parsers["upos"] = field_parsers["upostag"]
 
-    line = re.split(r"\t| {2,}", line)
+    line_split = re.split(r"\t| {2,}", line)
 
-    if len(line) == 1:
+    if len(line_split) == 1:
         raise ParseException("Invalid line format, line must contain either tabs or two spaces.")
 
     data = Token()
 
     for i, field in enumerate(fields):
         # Allow parsing CoNNL-U files with fewer columns
-        if i >= len(line):
+        if i >= len(line_split):
             break
 
         if field in field_parsers:
             try:
-                value = field_parsers[field](line, i)
+                value = field_parsers[field](line_split, i)
             except ParseException as e:
                 raise ParseException("Failed parsing field '{}': ".format(field) + str(e))
 
         else:
-            value = line[i]
+            value = line_split[i]
 
         data[str(field)] = value
 
     return data
 
-def parse_comment_line(line, metadata_parsers=None):
+def parse_comment_line(line: str,
+                       metadata_parsers: T.Optional[T.Dict[str, _MetadataParserType]] = None
+                       ) -> T.List[T.Tuple[str, T.Optional[str]]]:
     line = line.strip()
 
     if line[0] != '#':
@@ -162,17 +175,17 @@ def parse_comment_line(line, metadata_parsers=None):
 
     return [(str(key), value)]
 
-def parse_pair_value(value):
+def parse_pair_value(value: str) -> T.Tuple[str, T.Optional[str]]:
     key_maybe_value = value.split('=', 1)
     key = key_maybe_value[0].strip()
-    value = None if len(key_maybe_value) == 1 else key_maybe_value[1].strip()
+    value_new = None if len(key_maybe_value) == 1 else key_maybe_value[1].strip()
 
-    return key, value
+    return key, value_new
 
 
 INTEGER = re.compile(r"0|(\-?[1-9][0-9]*)")
 
-def parse_int_value(value):
+def parse_int_value(value: str) -> T.Optional[int]:
     if value == '_':
         return None
 
@@ -186,7 +199,7 @@ ID_SINGLE = re.compile(r"(?:0|[1-9][0-9]*)")
 ID_RANGE = re.compile(r"[1-9][0-9]*\-[1-9][0-9]*")
 ID_DOT_ID = re.compile(r"[0-9][0-9]*\.[1-9][0-9]*")
 
-def parse_id_value(value):
+def parse_id_value(value: str) -> T.Optional[_IdType]:
     if not value or value == '_':
         return None
 
@@ -194,8 +207,8 @@ def parse_id_value(value):
         return int(value)
 
     elif re.fullmatch(ID_RANGE, value):
-        from_, to = value.split("-")
-        from_, to = int(from_), int(to)
+        from_str, to_str = value.split("-")
+        from_, to = int(from_str), int(to_str)
         if to > from_:
             return (int(from_), "-", int(to))
 
@@ -209,7 +222,7 @@ ANY_ID = re.compile(ID_SINGLE.pattern + "|" + ID_RANGE.pattern + "|" + ID_DOT_ID
 DEPS_RE = re.compile("(" + ANY_ID.pattern + r")(:[^\d:_\-|][^:|]*)+")
 MULTI_DEPS_PATTERN = re.compile(r"{}(\|{})*".format(DEPS_RE.pattern, DEPS_RE.pattern))
 
-def parse_paired_list_value(value):
+def parse_paired_list_value(value: str) -> T.Union[T.Optional[str], T.List[T.Tuple[str, T.Optional[_IdType]]]]:
     if re.fullmatch(MULTI_DEPS_PATTERN, value):
         return [
             (part.split(":", 1)[1], parse_id_value(part.split(":")[0]))
@@ -218,31 +231,29 @@ def parse_paired_list_value(value):
 
     return parse_nullable_value(value)
 
-def parse_dict_value(value):
+def parse_dict_value(value: str) -> T.Optional[T.Dict[str, T.Optional[str]]]:
     if parse_nullable_value(value) is None:
         return None
 
-    return OrderedDict([
-        (part.split("=")[0], parse_nullable_value(part.split("=")[1]) if "=" in part else "")
+    return {
+        part.split("=")[0]: parse_nullable_value(part.split("=")[1]) if "=" in part else ""
         for part in value.split("|") if parse_nullable_value(part.split("=")[0]) is not None
-    ])
+    }
 
-def parse_nullable_value(value):
+def parse_nullable_value(value: str) -> T.Optional[str]:
     if not value or value == "_":
         return None
 
     return value
 
 # DEPRECATED: Mantain old paths until next major version
-
-def serialize(*args, **kwargs):
+def serialize(tokenlist: TokenList) -> str:
     from conllu.serializer import serialize as new_serialize
-    return new_serialize(*args, **kwargs)
+    return new_serialize(tokenlist)
 
-def serialize_field(*args, **kwargs):
+def serialize_field(field: T.Any) -> str:
     from conllu.serializer import serialize_field as new_serialize_field
-    return new_serialize_field(*args, **kwargs)
+    return new_serialize_field(field)
 
-def head_to_token(*args, **kwargs):
-    from conllu.models import TokenList
-    return TokenList.head_to_token(*args, **kwargs)
+def head_to_token(sentence: TokenList) -> T.Dict[int, T.List[Token]]:
+    return TokenList.head_to_token(sentence)
